"""
Data Fabric operations user interface
"""

import logging
import re
from nicegui import app, ui, run
import random
import string
import time

from ezlab.ezmeral.ezdf import client_install, cluster_install, xcluster

from ezlab.parameters import DF, DFAAS
from ezinfra.vms import prepare


logger = logging.getLogger("ezlab.ui.df")

@ui.refreshable
def add_remove_host(hosts: list):
    if hosts is None:
        hosts = []
    for host in hosts:
        with ui.row().classes("w-full items-center no-wrap"):
            ui.input(
                "Hostname",
                placeholder="Host shortnames, ie, ezvm1,ezvm2...",
            ).bind_value(host, "name")
            ui.input(
                "IP Address",
                placeholder="10.1.1.x",
            ).bind_value(host, "ip")


def get_doc(content: dict):

    random_eof = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))

    def indent(strings: str, spaces: int):
        return strings.strip().replace("\n", f"\n{' '*spaces}")

    settings = ""

    for key, value in content["settings"].items():
        settings += f"""
    #   {key}: {value if len(str(value)) < 80 else value[:77] + '...'}"""

    files = ""
    for filepath, filecontent in content["files"].items():
        files += f"""
    sudo cat <<{random_eof}>{filepath}
    {indent(filecontent,4)}
    {random_eof}
    """

    commands = ""
    for command in content["commands"]:
        commands += f"""
    {indent(command,4)}
    """

    doc = f"""
    #!/usr/bin/env bash

    # Ezlab Run {time.strftime('%Y%m%d%H%M%z')}
    
    # {content["task"]}

    # Settings
        {settings}

    # Write Files
        {files}
    
    # Run Commands
        {commands}

    # End of {content['task']}

"""

    with (
        ui.dialog().props("full-width") as dialog,
        ui.card().classes("w-full"),
    ):
        ui.code(doc, language="bash").classes("w-full text-wrap")
        with ui.row():
            ui.button(
                "Save (useless on app mode)",
                icon="save",
                on_click=lambda: ui.download(doc.encode(), "ezlab-runbook.sh"),
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=dialog.close,
            )
    dialog.open()


async def prepare_action():
    hosts = app.storage.general[DF]["hosts"]
    isDFaaS = app.storage.general[DF]["dfaas"]

    # deep check list[dict] for missing values (if any host's name or IP is missing)
    if not (
        len(hosts) > 0
        and all(
            [values for values in [all(host) for host in [h.values() for h in hosts]]]
        )
    ):
        ui.notify("Missing values", type="warning")
        return False

    logger.info("Starting configuration for DF")
    app.storage.user["busy"] = True
    for host in hosts:
        result = await run.io_bound(prepare,
            hostname=host["name"],
            hostip=host["ip"],
            settings=dict(app.storage.general["config"]),
            addhosts=[h for h in hosts if h["name"] != host["name"]],
            add_to_noproxy=(
                re.split(
                    ":|/", app.storage.general[DF].get("maprlocalrepo").split("://")[1]
                )[0]
                if app.storage.general[DF].get("maprrepoislocal", False)
                else None
            ),
            prepare_for=DFAAS if isDFaaS else DF,
            dryrun=app.storage.general["config"]["dryrun"],
        )
        if app.storage.general["config"]["dryrun"]:
            get_doc(result)
        elif result:
            logger.info("[ %s ] ready: %s", host['name'], result)
        else:
            ui.notify(f"[{host}] preparation failed!", type="warning")

    app.storage.user["busy"] = False


def prepare_menu():
    if not "hosts" in app.storage.general[DF].keys():
        app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]

    ui.checkbox("DF as a Service?").bind_value(app.storage.general[DF], "dfaas")

    with ui.row().classes("w-full items-center items-stretch") as prep_view:
        ui.label("Hosts")
        ui.space()
        with ui.row():

            def add_host():
                try:
                    app.storage.general[DF]["hosts"].append({"name": "", "ip": ""})
                except (KeyError, AttributeError):
                    app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]
                finally:
                    add_remove_host.refresh(app.storage.general[DF]["hosts"])

            def remove_host():
                if len(app.storage.general[DF]["hosts"]) > 0:
                    app.storage.general[DF]["hosts"].pop()
                    add_remove_host.refresh()

            ui.button(icon="add_box", color=None, on_click=add_host).props(
                "flat"
            )
            ui.button(icon="delete", color=None, on_click=remove_host).props(
                "flat"
            )

    # Hosts
    add_remove_host(app.storage.general[DF]["hosts"])


async def install_action(hosts_string: str):
    

    logger.info("Starting DF cluster installation")

    app.storage.user["busy"] = True

    result = await run.io_bound(cluster_install,
        hosts=hosts_string.split(","),
        settings=dict(app.storage.general["config"]),
        config=dict(app.storage.general[DF]),
        dryrun=app.storage.general["config"]["dryrun"],
    )

    if app.storage.general["config"]["dryrun"]:
        get_doc(result)

    elif result:
        ui.notify(f"Cluster is ready!", type="positive")

    else:
        ui.notify(f"Cluster creation failed!", type="negative")

    app.storage.user["busy"] = False


async def xcluster_action():

    logger.info("Starting DF cross-cluster setup: DISABLED")

    app.storage.user["busy"] = True

    if not await run.io_bound(
        xcluster,
        config=dict(app.storage.general["config"]),
        settings=dict(app.storage.general[DF]),
        dryrun=app.storage.general["config"]["dryrun"],
    ):
        ui.notify(f"Cluster creation failed!", type="negative")
        return False

    app.storage.user["busy"] = False
    return True


async def client_action():
    
    logger.info("Starting DF client configuration")

    app.storage.user["busy"] = True

    result = await run.io_bound(client_install,
        username=app.storage.general["config"]["username"],
        keyfile=app.storage.general["config"]["privatekeyfile"],
        config=dict(app.storage.general[DF]),
    )

    if not result:
        ui.notify("Client configuration failed", type="negative")

    app.storage.user["busy"] = False

    if result:
        ui.notify(f"Finished client configuration", type="positive")
        return True
    else:
        return False



import logging
from nicegui import app, ui, run
from ezlab.parameters import EZNODES, KVM, PVE
from ezinfra.vms import clone
from ezinfra import kvm, pve
from ezlab.utils import ezapp


logger = logging.getLogger("ezlab.ui.vms")

def pve_newvm_dialog():

    app.storage.user["busy"] = True

    templates = [
        x for x in pve.vms(ezapp.connection) if x["template"] and x["type"] == "qemu"
    ]
    # networks = pve.networks(ezapp.connection)
    storage = pve.storage(ezapp.connection)

    app.storage.user["busy"] = False

    # vm settings dialog
    with ui.dialog() as dialog, ui.card():
        template_selector = (
            ui.select(
                options=[x["name"] for x in templates],
                label="Template",
                # on_change=lambda e: set_template(e.value),
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "template")
        )
        storage_selector = (
            ui.select(
                options=[
                    x["storage"]
                    for x in storage
                    # if x["node"] == app.storage.user["template"]["node"]
                ],
                label="Storage",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "storage")
        )
        # network_selector = (
        #     ui.select(
        #         options=[
        #             x["sdn"]
        #             for x in networks
        #             # if x["node"] == app.storage.user["template"]["node"]
        #         ],
        #         label="Network",
        #     )
        #     .classes("w-full")
        #     .props("inline")
        #     .bind_value(app.storage.general[PVE], "network")
        # )

        bridge_selector = (
            ui.select(
                options=[
                    x["iface"]
                    for x in pve.bridges(ezapp.connection, app.storage.general[PVE]["template"])
                ],
                label="Bridge",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "bridge")
        )

        eznode_selector = (
            ui.select(
                options=[x["name"] for x in EZNODES],
                label="Node Type",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "eznode")
        )

        host_selector = (
            ui.input(label="VM Name", placeholder="ezvm")
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "hostname")
        )

        firstip_selector = (
            ui.input(
                label="First IP",
                placeholder=app.storage.general["config"]["gateway"],
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "firstip")
        )

        ui.button(
            "Create",
            on_click=lambda: dialog.submit(
                (
                    next(
                        (i for i in templates if i["name"] == template_selector.value),
                        None,
                    ),
                    next(
                        (i for i in storage if i["storage"] == storage_selector.value),
                        None,
                    ),
                    bridge_selector.value,
                    next(
                        (i for i in EZNODES if i["name"] == eznode_selector.value),
                        None,
                    ),
                    host_selector.value,
                    firstip_selector.value,
                )
            ),
        )

    return dialog


def kvm_newvm_dialog():
    # @ui.refreshable
    # def volume_selector():
    #     app.storage.user["busy"] = True
    #     vols = kvm.volumes(
    #         app.storage.general[KVM].get("pool", "default")
    #     )
    #     selector = (
    #         ui.select(
    #             options=vols,
    #             label="Base Image",
    #         )
    #         .classes("w-full")
    #         .props("inline")
    #         .bind_value(app.storage.general[KVM], "baseimg")
    #     )
    #     app.storage.user["busy"] = False
    #     return selector

    # select dialog
    with ui.dialog() as dialog, ui.card():

        app.storage.user["busy"] = True

        pools = kvm.pools()
        bridges = kvm.bridges()

        app.storage.user["busy"] = False

        pool = (
            ui.select(
                options=[p.name() for p in pools],
                label="Pool",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "pool")
        )

        baseimg = (
            ui.select(
                options=kvm.volumes(app.storage.general[KVM].get("pool", "default")),
                label="Base Image",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "baseimg")
        )
        

        bridge = (
            ui.select(
                options=bridges,
                label="Bridge",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "bridge")
        )

        eznode_selector = (
            ui.select(
                options=[x["name"] for x in EZNODES],
                label="Node Type",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "eznode")
        )

        host_selector = (
            ui.input(label="VM Name", placeholder="ezvm")
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "hostname")
        )

        firstip_selector = (
            ui.input(
                label="First IP",
                placeholder="10.1.1.21",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "firstip")
        )

        ui.button(
            "Create",
            on_click=lambda: dialog.submit(
                (
                    pool.value,
                    baseimg.value,
                    bridge.value,
                    next(
                        (i for i in EZNODES if i["name"] == eznode_selector.value),
                        None,
                    ),
                    host_selector.value,
                    firstip_selector.value,
                )
            ),
        )
    return dialog


async def new_vm_ui():

    if app.storage.general["target"]["hve"] == PVE:
        dialog = pve_newvm_dialog()
    elif app.storage.general["target"]["hve"] == KVM:
        dialog = kvm_newvm_dialog()
    else:
        ui.notify("not implemented")

    resources = await dialog

    if resources and all(resources):
        vmcount = resources[3]["count"]

        app.storage.user["busy"] = True

        for count in range(vmcount):
            try:
                await run.io_bound(clone,
                    target=app.storage.general["target"]["hve"],
                    resources=resources,
                    settings=dict(app.storage.general["config"]),
                    vm_number=(
                        0 if vmcount == 1 else count + 1
                    ),  # start from 1 not 0, use 0 for single vm
                    dryrun=app.storage.general["config"].get("dryrun", True),
                )
            except Exception as error:
                ui.notify(error, type="negative")

        app.storage.user["busy"] = False

    else:
        ui.notify("VM creation cancelled", type="info")



import asyncio
import base64
from datetime import datetime, timedelta
from ipaddress import IPv4Address, ip_address
import logging
import os
import shlex
import socket
import subprocess
import importlib.resources
import io
import json
import uuid

from fastapi.responses import StreamingResponse
from nicegui import app, ui


class ModuleContext:
    connection = None

    def __init__(self) -> None:
        pass


ezapp = ModuleContext()


class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        # change log format for UI
        self.setFormatter(
            logging.Formatter(
                "%(asctime)s:%(levelname)s: %(message)s",
                datefmt="%H:%M:%S",
            )
        )
        try:
            msg = self.format(record)
            self.element.push(msg)
        except Exception:
            self.handleError(record)


# Read the usage
README = importlib.resources.read_text("ezlab", "USAGE.md")


# validation for non-zero entry
def nonzero(val):
    return len(val) != 0


# wrapper to make sync calls async-like
def fire_and_forget(f):
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(
            None, f, *args, *[v for v in kwargs.values()]
        )

    return wrapped


# find app on system path
def find_app(name: str):
    """Return executable path for `name` on PATH or in CWD."""
    from shutil import which

    return which(cmd=name, path=f"{os.environ.get('PATH')}:{os.getcwd()}")


# execute app/script locally
def execute(cmd: str):
    proc = subprocess.Popen(
        args=shlex.split(cmd),
        stdout=subprocess.PIPE,
        universal_newlines=True,
        # env=sub_env,
    )

    while proc.stdout:
        output = proc.stdout.readline().strip()
        if output == "" and proc.poll() is not None:
            break
        if output:
            yield f">>> {output}"
    rc = proc.poll()


# Base64 conversion for secrets etc
def toB64(string: str):
    return base64.b64encode(string.encode("ascii")).decode("utf-8")


# IP address validation
def validate_ip(ip_string: str) -> bool:
    try:
        return isinstance(ip_address(ip_string), IPv4Address)
    except:
        return False


# reliably get FQDN for given IP
def get_fqdn(ip_string: str) -> str:
    fqdn, _, _ = socket.gethostbyaddr(ip_string)
    return fqdn


def is_file(file):
    return os.path.isfile(os.path.abspath(file))


def dt_from_iso(timestring: str):
    # Workaround since received timestring with " AM"/" PM" suffix is not parsed properly
    isPM = " PM" in timestring
    dt = datetime.strptime(
        timestring.replace(" AM", "").replace(" PM", ""), "%Y-%m-%dT%H:%M:%S.%f%z"
    )
    return dt + timedelta(hours=12) if isPM else dt


def create_ci_iso(files: dict, destfile: str):
    from io import BytesIO
    import pycdlib

    metadata = files["meta-data"]
    userdata = files["user-data"]
    netconf = files["network-config"]

    iso = pycdlib.PyCdlib()
    iso.new(
        interchange_level=3,
        joliet=True,
        sys_ident="LINUX",
        rock_ridge="1.09",
        vol_ident="cidata",
    )

    metadata_file = BytesIO(metadata.encode())
    metadata_file.seek(0)
    iso.add_fp(
        metadata_file,
        len(metadata),
        "/METADATA.;1",
        rr_name="meta-data",
        joliet_path="/meta-data",
    )

    userdata_file = BytesIO(userdata.encode())
    iso.add_fp(
        userdata_file,
        len(userdata),
        "/USERDATA.;1",
        rr_name="user-data",
        joliet_path="/user-data",
    )

    netconf_file = BytesIO(netconf.encode())
    netconf_file.seek(0)
    iso.add_fp(
        netconf_file,
        len(netconf),
        "/NETWORK_.;1",
        rr_name="network-config",
        joliet_path="/network-config",
    )

    iso.write(destfile)
    iso.close()
    userdata_file.close()
    metadata_file.close()
    netconf_file.close()


# Define a download path for config downloads
CONFIG_DL_PATH = f"/config/{uuid.uuid4()}.json"

@app.get(CONFIG_DL_PATH)
def download():
    string_io = io.StringIO(
        json.dumps(app.storage.general)
    )  # create a file-like object from the string
    headers = {"Content-Disposition": "attachment; filename=config.json"}
    return StreamingResponse(string_io, media_type="text/plain", headers=headers)


APP_STATUS_CONTAINER = []

@ui.refreshable
def status_container():
    with ui.row():
        for stat in APP_STATUS_CONTAINER:
            item, dest = stat
            ui.link(text=item, target=dest, new_tab=True).classes("text-emerald-400")
            ui.space()
